//
// Created by rhobincu on 5/23/23.
//

#pragma once

#include <queue>
#include <mutex>
#include <condition_variable>

namespace dcae::pclp2 {
    template <typename T> class AsynchronousBlockingQueue {
        std::queue<T> mData;
        size_t mMaxLength;
        std::mutex mMutex;
        std::condition_variable mConditionVariable;
    public:
        AsynchronousBlockingQueue(size_t maxLength) : mMaxLength(maxLength) {

        }

        void push(const T & newElement) {
            std::unique_lock<std::mutex> lock(mMutex);
            while(mData.size() >= mMaxLength) {
                mConditionVariable.wait(lock);
            }
            mData.push(newElement);
            mConditionVariable.notify_all();
        }

        T pop() {
            std::unique_lock<std::mutex> lock(mMutex);
            while(mData.empty()) {
                mConditionVariable.wait(lock);
            }
            T toReturn = mData.front();
            mData.pop();
            mConditionVariable.notify_all();
            return toReturn;
        }

        T peek() {
            std::unique_lock<std::mutex> lock(mMutex);
            while(mData.empty()) {
                mConditionVariable.wait(lock);
            }
            return mData.front();
        }

        size_t size() {
            std::unique_lock<std::mutex> lock(mMutex);
            return mData.size();
        }

    };
}
