//
// Created by rhobincu on 4/4/23.
//

#pragma once

#include <iostream>

namespace etti::pclp2 {
    /**
     * @brief Class template that represents an auto-resizable vector.
     * @tparam T The type of elements this vector holds.
     */
    template<typename T=int> class Vector {
        T *mData;
        size_t mCapacity;
        size_t mSize;
        static uint32_t sVectorObjects;

        /* Doubles the existing capacity */
        void increaseArrayCapacity() {
            mCapacity *= 2;
            T *newBuffer = new T[mCapacity];
            for (size_t i = 0; i < mSize; i++) newBuffer[i] = mData[i];
            delete[] mData;
            mData = newBuffer;
        }

    public:
        Vector(size_t capacity) : mData(new T[capacity]), mCapacity(capacity), mSize(0) {
            std::cout << "S-a apelat constructorul cu arg de tip size_t" << std::endl;
            sVectorObjects++;
        }

        Vector() : Vector(100) {
            std::cout << "S-a apelat constructorul fara argument" << std::endl;
        }

        Vector(const Vector &source) : mSize(source.mSize), mCapacity(source.mCapacity),
                mData(new T[source.mCapacity]) {
            for (size_t i = 0; i < mSize; i++) mData[i] = source.mData[i];
            std::cout << "S-a apelat constructorul de copiere" << std::endl;
            sVectorObjects++;
        }

        Vector(Vector &&source) : mSize(source.mSize), mCapacity(source.mCapacity), mData(source.mData) {
            source.mData = nullptr;
            std::cout << "S-a apelat constructorul de mutare" << std::endl;
        }

        static uint32_t getObjectCount() {
            return sVectorObjects;
        }

        Vector &operator=(const Vector &source) {
            if (&source == this) return *this;
            if (mCapacity < source.mCapacity) {
                delete[] mData;
                mData = new T[source.mCapacity];
                mCapacity = source.mCapacity;
            }
            mSize = source.mSize;
            for (size_t i = 0; i < mSize; i++) mData[i] = source.mData[i];
            std::cout << "S-a apelat operatorul de atribuire" << std::endl;
            return *this;
        }

        Vector &operator=(Vector &&source) {
            delete[] mData;
            mSize = source.mSize;
            mCapacity = source.mCapacity;
            mData = source.mData;
            source.mData = nullptr;
            std::cout << "S-a apelat operatorul de mutare" << std::endl;
            return *this;
        }

        size_t capacity() const {
            return mCapacity;
        }

        size_t size() const {
            return mSize;
        }

        virtual void push_back(T newElement) {
            if (mSize == mCapacity) {
                increaseArrayCapacity();
            }
            mData[mSize++] = newElement;
        }

        T &at(size_t index) {
            if (index >= mSize) {
                throw std::out_of_range(
                        "Invalid index " + std::to_string(index) + "; max size is: " + std::to_string(mSize));
            }
            return mData[index];
        }

        T at(size_t index) const {
            if (index >= mSize) {
                throw std::out_of_range(
                        "Invalid index " + std::to_string(index) + "; max size is: " + std::to_string(mSize));
            }

            return mData[index];
        }

        T &operator[](size_t index) {
            return mData[index];
        }

        T operator[](size_t index) const {
            return mData[index];
        }

        T &front() {
            return mData[0];
        }

        T front() const {
            return mData[0];
        }

        T &back() {
            return mData[mSize - 1];
        }

        T back() const {
            return mData[mSize - 1];
        }

        virtual ~Vector() {
            delete[] mData;
            std::cout << "S-a apelat destructorul" << std::endl;
            sVectorObjects--;
        }
    };

    template<typename T> uint32_t etti::pclp2::Vector<T>::sVectorObjects = 0;
}
