#include <iostream>
#include <vector>
#include <thread>
#include <mutex>
#include <atomic>
#include <future>
#include "etti/pclp2/Vector.hpp"
#include "etti/pclp2/Shapes.hpp"
#include "AsynchronousBlockingQueue.hpp"

int sumAll(const etti::pclp2::Vector<int> &array) {
    int sum = 0;
    for (int i = 0; i < array.size(); i++) sum += array[i];
    return sum;
}


class ShapeVector : public etti::pclp2::Vector<Shape *> {
    int *mField;
public:
    ShapeVector() {
        mField = new int[20];
    }

    double computeWholeArea() {
        double area = 0;
        for (int i = 0; i < size(); i++) area += at(i)->getArea();
        return area;
    }

    ~ShapeVector() override {
        delete[] mField;
        std::cout << "Called destructor for ShapeVector" << std::endl;
    }
};

std::mutex counterMutex;
std::atomic<int32_t> gCounter = 0;

/**
 * @brief Prints the given message, periodically, with the given delay in ms
 * @param message message to display
 * @param delay delay between each prints, in ms
 * @param count how many times the message is printed
 */
void printMessage(const std::string & message, uint32_t delay, uint32_t count) {
    while(count-- > 0) {
        std::cout << message << std::endl;
        std::this_thread::sleep_for(std::chrono::milliseconds(delay));
    }
}

void modifyVariable(int32_t increment) {
    //std::unique_lock<std::mutex> myLock(counterMutex);
//    counterMutex.lock(); // zona de cod dintre lock si unlock este executata de un singur thread la un moment dat (zona sincronizata)
    if(abs(increment) >= 100) {
//        counterMutex.unlock();
        throw std::runtime_error("Increment too large!");
    }
    gCounter += increment; // load, add, store -> 3 instructiuni => race condition
//    counterMutex.unlock();
}

void incrementInLoop(int32_t increment, uint32_t delay, uint32_t count) {
    try {
        while (count-- > 0) {
            std::this_thread::sleep_for(std::chrono::nanoseconds(delay));
            modifyVariable(increment);
        }
    }catch(std::runtime_error & ex) {
        std::cerr << "Error: " << ex.what() << std::endl;
    }
}

size_t valuesToAdd = 20;

void produce(dcae::pclp2::AsynchronousBlockingQueue<int> * queue) {
    for(size_t i = 0; i < valuesToAdd; i++) {
        std::cout << "producing value " << i << std::endl;
        queue->push(i);
        std::this_thread::sleep_for(std::chrono::milliseconds(100));
    }
}

void consume(dcae::pclp2::AsynchronousBlockingQueue<int> * queue) {
    for(size_t i = 0; i < valuesToAdd; i++) {
        std::cout << "CONSUMING VALUE " << queue->pop() << std::endl;
        std::this_thread::sleep_for(std::chrono::milliseconds(200));
    }
}

double functionThatHasALotOfWork(int a, int b, int c) {
    std::this_thread::sleep_for(std::chrono::milliseconds(1000));
    return (a + b + c) / 3.0;
}

int main() {
    //    Vector v1;
    //    Vector v2(20);
    //    std::cout << "Capacity of v1 is " << v1.capacity() << " and size is: " << v1.size() << std::endl;
    //    std::cout << "Capacity of v2 is " << v2.capacity() << " and size is: " << v2.size() << std::endl;
    //
    //    v1.push_back(1);
    //    v2.push_back(2);
    //    v2.push_back(3);
    //
    //    std::cout << "Capacity of v1 is " << v1.capacity() << " and size is: " << v1.size() << std::endl;
    //    std::cout << "Capacity of v2 is " << v2.capacity() << " and size is: " << v2.size() << std::endl;
    //
    //    std::cout << "v1[0] = " << v1.at(0) << std::endl;
    //    std::cout << "v2[0] = " << v2.at(0) << std::endl;
    //    std::cout << "v2[1] = " << v2.at(1) << std::endl;
    //
    //    Vector smallVector(2);
    //    smallVector.push_back(1);
    //    smallVector.push_back(2);
    //    smallVector.push_back(5);
    //    smallVector.push_back(-20);
    //    smallVector.push_back(71);
    //
    //    std::cout << "Capacity of smallVector is " << smallVector.capacity() << " and size is: " << smallVector.size()
    //              << std::endl;
    //    std::cout << "Elements: ";
    //    smallVector.at(0) = -12; // smallVector[0] = -12;
    //    for (size_t i = 0; i < smallVector.size(); i++) std::cout << smallVector.at(i) << ", ";
    //    std::cout << std::endl;
    //
    //    Vector copy = smallVector;



    //    etti::pclp2::Vector<int> v1;
    //
    //    v1 = etti::pclp2::Vector<int>(1000);
    //    auto v2 = etti::pclp2::Vector<double>(1000);
    //
    //
    //    int n = 6;
    //    etti::pclp2::Array<double, 2> arr1;
    //
    //    arr1[2] = 1.3;

    //    ro::upb::etti::StudentWithGrade s2("Gigi", 20, 9.99);
    //    s2.incAge();
    //    s2.grade = 4.0;
    //    std::cout << s2.getName() << " are acum " << s2.getAge() << " ani!" << std::endl;
    //
    //    ro::upb::etti::Student s3("Ion", 20);
    //    s3.incAge();
    //    std::cout << s3.getName() << " are acum " << s3.getAge() << " ani!" << std::endl;

    //    using ro::upb::etti::StudentWithGrade;
    //    using ro::upb::etti::Student;
    //
    //    Student * gigi = new Student("Gigi", 20);
    //    StudentWithGrade * ion = new StudentWithGrade("Ion", 20, 7.0);
    //    Student * ana = new StudentWithGrade("Ana", 20, 8.0);
    //
    //    gigi->incAge();
    //    ion->incAge();
    //    ana->incAge();

    //ShapeVector shapes;
//    etti::pclp2::Vector<Shape *> *shapes = new ShapeVector();
//
//    shapes->push_back(new Square(10));
//    shapes->push_back(new Circle(3));
//    shapes->push_back(new Rectangle(3, 4));
//    shapes->push_back(new Triangle(3, 4, 5));
//
//    try {
//        std::cout << "Suprafata formei de pe pozitia -2 este: " << shapes->at(-30)->getArea() << std::endl;
//        std::cout << "Dupa apelul dubios al lui .at()" << std::endl;
//    } catch (std::out_of_range &ex) {
//        std::cerr << "Exceptie out of range!!!: " << ex.what() << std::endl;
//    } catch (std::logic_error &ex) {
//        std::cerr << "Exceptie: " << ex.what() << std::endl;
//    } catch (std::exception &ex) {
//        std::cout << ex.what() << std::endl;
//    } catch (...) {
//        std::cout << "Exceptie generica!... (nu de tip std::exception)..." << std::endl;
//    }
//
//    ShapeVector *castShapes = dynamic_cast<ShapeVector *>(shapes);
//    if (castShapes != nullptr) {
//        std::cout << "Suprafata totala este: " << castShapes->computeWholeArea() << std::endl;
//    }
//
//    for (int i = 0; i < shapes->size(); i++) delete shapes->at(i);
//
//    etti::pclp2::Vector<Shape*> v2;
//
//    std::cout << "We have " << etti::pclp2::Vector<Shape *>::getObjectCount() << " objects of type Vector<Shape*> " << std::endl;
//
//    delete shapes;
//
//    std::cout << "We have " << etti::pclp2::Vector<Shape *>::getObjectCount() << " objects of type Vector<Shape*> " << std::endl;

//    std::thread goodMorningThread(printMessage, "Good morning world!", 700, 3);
//    printMessage("Hello world!", 500, 4);
//    //printMessage("Good morning world!", 700);
//    goodMorningThread.join();
    //std::cout << "End of program!" << std::endl;

//    std::thread incThread(incrementInLoop, 1, 0, 100000000);
//    std::thread decThread(incrementInLoop, -1, 0, 100000000);
//
//    incThread.join();
//    decThread.join();

//    std::cout << "Global var is: " << gCounter << std::endl;

//    dcae::pclp2::AsynchronousBlockingQueue<int> queue(5);
//
//    std::thread producer(produce, &queue);
//    std::thread consumer(consume, &queue);
//
//    producer.join();
//    consumer.join();

    std::future<double> myResult = std::async(std::launch::async, functionThatHasALotOfWork, 1, 2, 3);
    std::cout << "I'm doing some other stuff in main!" << std::endl;
    std::cout << "I'm doing some other stuff in main!" << std::endl;
    std::this_thread::sleep_for(std::chrono::milliseconds(900));
    std::cout << "I'm doing some other stuff in main!" << std::endl;
    std::cout << "I'm doing some other stuff in main!" << std::endl;

    std::cout << "Result is: " << myResult.get() << std::endl;


    return 0;
}
